@@ -355,6 +355,7 @@ gnome_vfs_module_callback_full_authentication_flags_get_type (void)
       { GNOME_VFS_MODULE_CALLBACK_FULL_AUTHENTICATION_NEED_USERNAME, "GNOME_VFS_MODULE_CALLBACK_FULL_AUTHENTICATION_NEED_USERNAME", "need-username" },
       { GNOME_VFS_MODULE_CALLBACK_FULL_AUTHENTICATION_NEED_DOMAIN, "GNOME_VFS_MODULE_CALLBACK_FULL_AUTHENTICATION_NEED_DOMAIN", "need-domain" },
       { GNOME_VFS_MODULE_CALLBACK_FULL_AUTHENTICATION_SAVING_SUPPORTED, "GNOME_VFS_MODULE_CALLBACK_FULL_AUTHENTICATION_SAVING_SUPPORTED", "saving-supported" },
+      { GNOME_VFS_MODULE_CALLBACK_FULL_AUTHENTICATION_ANON_SUPPORTED, "GNOME_VFS_MODULE_CALLBACK_FULL_AUTHENTICATION_ANON_SUPPORTED", "anon-supported" },
       { 0, NULL, NULL }
     };
     etype = g_flags_register_static ("GnomeVFSModuleCallbackFullAuthenticationFlags", values);
@@ -362,6 +363,20 @@ gnome_vfs_module_callback_full_authentication_flags_get_type (void)
   return etype;
 }
 
+GType
+gnome_vfs_module_callback_full_authentication_out_flags_get_type (void)
+{
+  static GType etype = 0;
+  if (etype == 0) {
+    static const GFlagsValue values[] = {
+      { GNOME_VFS_MODULE_CALLBACK_FULL_AUTHENTICATION_OUT_ANON_SELECTED, "GNOME_VFS_MODULE_CALLBACK_FULL_AUTHENTICATION_OUT_ANON_SELECTED", "selected" },
+      { 0, NULL, NULL }
+    };
+    etype = g_flags_register_static ("GnomeVFSModuleCallbackFullAuthenticationOutFlags", values);
+  }
+  return etype;
+}
+
 
 /* enumerations from "gnome-vfs-utils.h" */
 GType
