@@ -51,6 +51,7 @@
 
 #include <config.h>
 
+#include <libgnomevfs/gnome-vfs-i18n.h>
 #include <libgnomevfs/gnome-vfs-context.h>
 #include <libgnomevfs/gnome-vfs-method.h>
 #include <libgnomevfs/gnome-vfs-module.h>
@@ -844,24 +845,31 @@ iobuf_send_string_request_with_file_info (int                      fd,
 
 
 
-static gboolean
-invoke_full_auth (const GnomeVFSURI *uri,
-		  gboolean done_auth,
-		  const char *password_line,
-		  char **password_out)
+static char*
+get_user_from_uri_or_password_line (const GnomeVFSURI *uri,
+				    const char *password_line)
 {
-	GnomeVFSModuleCallbackFullAuthenticationIn in_args;
-	GnomeVFSModuleCallbackFullAuthenticationOut out_args;
-	gboolean invoked;
-	char *user, *chr, *ptr;
-	char *object;
-	gboolean passphrase;
+	char *chr, *user = NULL;
 
-	passphrase = g_str_has_prefix (password_line, "Enter passphrase for key");
+	if (!g_str_has_prefix (password_line, "Enter passphrase for key")) {
+		chr = strchr (password_line, '@');
 
-	user = NULL;
-	object = NULL;
-	if (passphrase) {
+		if (chr != NULL) {
+			user = g_strndup (password_line, chr - password_line);
+		}
+	}
+	if (user == NULL) {
+		user = g_strdup ((char *)gnome_vfs_uri_get_user_name (uri));
+	}
+	return user;
+}
+
+static char*
+get_object_from_password_line (const char *password_line)
+{
+	char *chr, *ptr, *object = NULL;
+
+	if (g_str_has_prefix (password_line, "Enter passphrase for key")) {
 		ptr = strchr (password_line, '\'');
 		if (ptr != NULL) {
 			ptr += 1;
@@ -872,13 +880,89 @@ invoke_full_auth (const GnomeVFSURI *uri,
 				object = g_strdup (ptr);
 			}
 		}
-	} else {
-		chr = strchr (password_line, '@');
-
-		if (chr != NULL) {
-			user = g_strndup (password_line, chr - password_line);
-		}
 	}
+	return object;
+}
+
+static char*
+get_server_from_uri_or_password_line (const GnomeVFSURI *uri,
+				      const char *password_line)
+{
+	if (!g_str_has_prefix (password_line, "Enter passphrase for key")) {
+		return g_strdup ( (char *)gnome_vfs_uri_get_host_name (uri));
+	}
+	return NULL;
+}
+
+static char*
+get_authtype_from_password_line (const char *password_line)
+{
+	if (g_str_has_prefix (password_line, "Enter passphrase for key")) {
+		return g_strdup ("publickey");
+	} 
+	return g_strdup ("password");
+}
+
+
+
+static gboolean
+invoke_fill_auth (const GnomeVFSURI *uri,
+		  const char *password_line,
+		  char **password_out)
+{
+	gboolean invoked;
+	GnomeVFSModuleCallbackFillAuthenticationIn in_args;
+	GnomeVFSModuleCallbackFillAuthenticationOut out_args;
+
+
+	memset (&in_args, 0, sizeof (in_args));
+	in_args.protocol = "sftp";
+	
+	in_args.uri = gnome_vfs_uri_to_string (uri, 0);
+	in_args.object = get_object_from_password_line (password_line);
+	in_args.authtype = get_authtype_from_password_line (password_line);
+	in_args.domain = NULL;
+	in_args.port = gnome_vfs_uri_get_host_port (uri);
+	in_args.server = get_server_from_uri_or_password_line (uri, password_line);
+	in_args.username = get_user_from_uri_or_password_line (uri, password_line);
+	memset (&out_args, 0, sizeof (out_args));
+
+	invoked = gnome_vfs_module_callback_invoke
+		  (GNOME_VFS_MODULE_CALLBACK_FILL_AUTHENTICATION,
+		  &in_args, sizeof (in_args),
+		  &out_args, sizeof (out_args));
+	if (invoked && out_args.valid) {
+		*password_out = g_strdup (out_args.password);
+		g_free (out_args.username);
+		g_free (out_args.domain);
+		g_free (out_args.password);
+	} else {
+		*password_out = NULL;
+	}
+
+	g_free (in_args.uri);
+	g_free (in_args.username);
+	g_free (in_args.object);
+	g_free (in_args.server);
+	g_free (in_args.authtype);
+
+	return invoked && out_args.valid;
+}
+
+static gboolean
+invoke_full_auth (const GnomeVFSURI *uri,
+		  gboolean done_auth,
+		  const char *password_line,
+		  char **password_out,
+		  char **keyring_out,
+		  char **user_out,
+		  char **object_out,
+		  char **authtype_out,
+		  gboolean *save_password_out)
+{
+	GnomeVFSModuleCallbackFullAuthenticationIn in_args;
+	GnomeVFSModuleCallbackFullAuthenticationOut out_args;
+	gboolean invoked;
 
 	memset (&in_args, 0, sizeof (in_args));
 	in_args.uri = gnome_vfs_uri_to_string (uri, 0);
@@ -886,22 +970,13 @@ invoke_full_auth (const GnomeVFSURI *uri,
 	if (done_auth) {
 		in_args.flags |= GNOME_VFS_MODULE_CALLBACK_FULL_AUTHENTICATION_PREVIOUS_ATTEMPT_FAILED;
 	}
-	/* TODO: Support username too. A bit harder */
-
 	in_args.protocol = "sftp";
-	if (!passphrase) {
-		in_args.server = (char *)gnome_vfs_uri_get_host_name (uri);
-		if (user != NULL) {
-			in_args.username = user;
-		} else {
-			in_args.username = (char *)gnome_vfs_uri_get_user_name (uri);
-		}
-	}
-	in_args.object = object;
-	in_args.authtype = passphrase ? "publickey": "password";
-	
+	in_args.object = get_object_from_password_line (password_line);
+	in_args.authtype = get_authtype_from_password_line (password_line);
 	in_args.domain = NULL;
 	in_args.port = gnome_vfs_uri_get_host_port (uri);
+	in_args.server = get_server_from_uri_or_password_line (uri, password_line);
+	in_args.username = get_user_from_uri_or_password_line (uri, password_line);
 
 	memset (&out_args, 0, sizeof (out_args));
 
@@ -909,24 +984,61 @@ invoke_full_auth (const GnomeVFSURI *uri,
 		(GNOME_VFS_MODULE_CALLBACK_FULL_AUTHENTICATION,
 		 &in_args, sizeof (in_args),
 		 &out_args, sizeof (out_args));
-
 	if (invoked && !out_args.abort_auth) {
+		if (out_args.save_password) {
+			*keyring_out = g_strdup (out_args.keyring);
+			*user_out = get_user_from_uri_or_password_line (uri, password_line);
+			*object_out = get_object_from_password_line (password_line);
+			*authtype_out = get_authtype_from_password_line (password_line);
+		}
 		*password_out = g_strdup (out_args.password);
+		*save_password_out = out_args.save_password;
+		g_free (out_args.username);
+		g_free (out_args.domain);
+		g_free (out_args.password);
+		g_free (out_args.keyring);
 	} else {
 		*password_out = NULL;
 	}
-	g_free (out_args.username);
-	g_free (out_args.domain);
-	g_free (out_args.password);
-	g_free (out_args.keyring);
 
 	g_free (in_args.uri);
-	g_free (user);
-	g_free (object);
+	g_free (in_args.username);
+	g_free (in_args.object);
+	g_free (in_args.server);
+	g_free (in_args.authtype);
 
 	return invoked && !out_args.abort_auth;
 }
 
+static void
+invoke_save_auth (const GnomeVFSURI *uri,
+	   char *keyring,
+	   char *user,
+	   char *object,
+	   char *authtype,
+	   char *password)
+{
+	GnomeVFSModuleCallbackSaveAuthenticationIn save_in_args;
+	GnomeVFSModuleCallbackSaveAuthenticationOut save_out_args;
+
+	memset (&save_in_args, 0, sizeof (save_in_args));
+	save_in_args.uri = gnome_vfs_uri_to_string (uri, 0);
+	save_in_args.server = (char *)gnome_vfs_uri_get_host_name (uri);
+	save_in_args.port = gnome_vfs_uri_get_host_port (uri);
+	save_in_args.protocol = "sftp";
+	save_in_args.keyring = keyring;
+	save_in_args.username = user;
+	save_in_args.object = object;
+	save_in_args.authtype = authtype;
+	save_in_args.password = password;
+
+	memset (&save_out_args, 0, sizeof (save_out_args));
+	gnome_vfs_module_callback_invoke (GNOME_VFS_MODULE_CALLBACK_SAVE_AUTHENTICATION,
+					  &save_in_args, sizeof (save_in_args),
+					  &save_out_args, sizeof (save_out_args));
+	g_free (save_in_args.uri);
+}
+	
 
 /* Derived from OpenSSH, sftp.c:main */
 
@@ -941,14 +1053,24 @@ sftp_connect (SftpConnection **connection, const GnomeVFSURI *uri)
 	const gchar    *user_name;
 	gint            port;
 	guint           last_arg, i;
+	gboolean        full_auth;
 	gboolean        done_auth;
+	gboolean	save_password;
 	Buffer          msg;
 	gchar           type;
+	char *password;
+	char *keyring;
+	char *user;
+	char *object;
+	char *authtype;
 
 	GError         *error = NULL;
 
 	gchar          *args[20]; /* Enough for now, extend if you add more args */
-
+	gboolean invoked; 		  
+	GnomeVFSModuleCallbackQuestionIn in_args; 
+	GnomeVFSModuleCallbackQuestionOut out_args;
+	
 	DEBUG (gchar *tmp);
 
 	/* Fill in the first few args */
@@ -1042,6 +1164,7 @@ sftp_connect (SftpConnection **connection, const GnomeVFSURI *uri)
 					g_io_channel_get_flags (tty_channel) | G_IO_FLAG_NONBLOCK, NULL);
 	}
 	done_auth = FALSE;
+	full_auth = FALSE;
 	while (tty_fd != -1) {
 		fd_set ifds;
 		struct timeval tv;
@@ -1049,8 +1172,13 @@ sftp_connect (SftpConnection **connection, const GnomeVFSURI *uri)
 		GIOStatus io_status;
 		char buffer[1024];
 		gsize len;
-		char *password;
-
+		char *choices[3];
+		char *pos;
+		char *startpos;
+		char *endpos;
+		char hostname[256];
+		char fingerprint[256];
+		
 		FD_ZERO (&ifds);
 		FD_SET (in_fd, &ifds);
 		FD_SET (tty_fd, &ifds);
@@ -1078,7 +1206,13 @@ sftp_connect (SftpConnection **connection, const GnomeVFSURI *uri)
 			if (g_str_has_suffix (buffer, "password: ") ||
 			    g_str_has_suffix (buffer, "Password: ") ||
 			    g_str_has_prefix (buffer, "Enter passphrase for key")) {
-				if (invoke_full_auth (uri, done_auth, buffer, &password) && password != NULL) {
+				if (!done_auth && invoke_fill_auth (uri, buffer, &password) && password != NULL) {
+					g_io_channel_write_chars (tty_channel, password, -1, &len, NULL);
+					g_io_channel_write_chars (tty_channel, "\n", 1, &len, NULL);
+					g_io_channel_flush (tty_channel, NULL);
+				} else if (invoke_full_auth (uri, done_auth, buffer, &password, &keyring, 
+							     &user, &object, &authtype, &save_password) && password != NULL) {
+					full_auth = TRUE;
 					g_io_channel_write_chars (tty_channel, password, -1, &len, NULL);
 					g_io_channel_write_chars (tty_channel, "\n", 1, &len, NULL);
 					g_io_channel_flush (tty_channel, NULL);
@@ -1087,12 +1221,83 @@ sftp_connect (SftpConnection **connection, const GnomeVFSURI *uri)
 					goto bail;
 				}
 				done_auth = TRUE;
-			} else if (g_str_has_prefix (buffer, "The authenticity of host")) {
-				/* FIXME: This should do a callback asking the user if the host id is ok */
-				/* For now we just fail */
-				res = GNOME_VFS_ERROR_ACCESS_DENIED;
-				goto bail;
-                        }
+			} else if (g_str_has_prefix (buffer, "The authenticity of host '")) {
+
+				pos = strchr (&buffer[26], '\'');
+				if (pos == NULL) {
+					res = GNOME_VFS_ERROR_GENERIC;
+					goto bail;
+				}
+
+				if (pos - (&buffer[26]) < 255) {
+					strncpy (hostname, &buffer[26], pos - (&buffer[26]));
+					hostname[pos-(&buffer[26])]='\0';
+				} else {
+					strncpy (hostname, &buffer[26], 255);
+					hostname[255]='\0';
+				}
+				
+				startpos = strstr (pos, "RSA key fingerprint is ");
+				if (startpos == NULL) {
+					res = GNOME_VFS_ERROR_GENERIC;
+					goto bail;
+				}
+			
+				startpos = startpos + 23;
+				endpos = strchr (startpos, '.');
+				if (endpos == NULL) {
+					res = GNOME_VFS_ERROR_GENERIC;
+					goto bail;
+				}
+				
+				if (endpos - startpos < 255) {
+					strncpy (fingerprint, startpos, endpos - startpos);
+					fingerprint[endpos - startpos]='\0';
+				} else {
+					strncpy (fingerprint, startpos, 255);
+					fingerprint[255]='\0';
+				}
+				
+				in_args.primary_message = g_strdup_printf (_("The identity of the remote computer (%s) is unknown."), hostname);
+				in_args.secondary_message = g_strdup_printf (_("This happens when you log in to a computer the first time.\n\n"
+									       "The identity sent by the remote computer is %s. "
+									       "If you want to be absolutely sure it is safe to continue, contact the system administrator."), fingerprint);
+
+				in_args.choices = choices;
+				in_args.choices[0] = _("Log In Anyway");
+				in_args.choices[1] = _("Cancel Login");
+				in_args.choices[2] = NULL;
+								
+				invoked = gnome_vfs_module_callback_invoke
+					(GNOME_VFS_MODULE_CALLBACK_QUESTION,
+					 &in_args, sizeof (in_args),
+					 &out_args, sizeof (out_args));
+				
+				if (invoked) {
+					if (out_args.answer == 0) {
+						g_io_channel_write_chars (tty_channel, "yes\n", -1, &len, NULL);
+					} else {
+						g_io_channel_write_chars (tty_channel, "no\n", -1, &len, NULL);
+						g_free (in_args.primary_message);
+						g_free (in_args.secondary_message);
+						res = GNOME_VFS_ERROR_ACCESS_DENIED;
+						goto bail;
+					}
+					g_io_channel_flush (tty_channel, NULL);
+					buffer[0]='\0';
+				} else {
+					g_io_channel_write_chars (tty_channel, "no\n", -1, &len, NULL);
+					g_io_channel_flush (tty_channel, NULL);
+					buffer[0]='\0';
+					g_free (in_args.primary_message);
+					g_free (in_args.secondary_message);
+					res = GNOME_VFS_ERROR_ACCESS_DENIED;
+					goto bail;
+				}
+				g_free (in_args.primary_message);
+				g_free (in_args.secondary_message);
+			}
+			
 		}
 	}
 
@@ -1113,6 +1318,9 @@ sftp_connect (SftpConnection **connection, const GnomeVFSURI *uri)
 		res = GNOME_VFS_ERROR_PROTOCOL_ERROR;
 	} else {
 		/* Everything's A-OK. Set up the connection and go */
+		if (full_auth == TRUE && save_password == TRUE) {
+			invoke_save_auth (uri, keyring, user, object, authtype, password);
+		}
 
 		if (!g_thread_supported ()) g_thread_init (NULL);
 
@@ -1136,9 +1344,14 @@ sftp_connect (SftpConnection **connection, const GnomeVFSURI *uri)
 	}
 
  bail:
-
 	buffer_free (&msg);
 
+	g_free (password);
+	g_free (keyring);
+	g_free (user);
+	g_free (object);
+	g_free (authtype);
+
 	if (res != GNOME_VFS_OK) {
 		close (in_fd);
 		close (out_fd);
@@ -1174,10 +1387,7 @@ sftp_get_connection (SftpConnection **connection, const GnomeVFSURI *uri)
 		user_name = g_get_user_name ();
 
 	if (host_name == NULL)
-	{
-		res = GNOME_VFS_ERROR_HOST_NOT_FOUND;
-		goto bail;
-	}
+		return GNOME_VFS_ERROR_HOST_NOT_FOUND;
 
 	hash_name = g_strconcat (user_name, "@", host_name, NULL);
 
@@ -1193,11 +1403,7 @@ sftp_get_connection (SftpConnection **connection, const GnomeVFSURI *uri)
 		res = sftp_connect (connection, uri);
 
 		if (res == GNOME_VFS_OK) {
-			if (*connection == NULL)
-			{
-				res = GNOME_VFS_ERROR_INTERNAL;
-				goto bail;
-			}
+			g_return_val_if_fail (*connection != NULL, GNOME_VFS_ERROR_INTERNAL);
 			g_mutex_lock ((*connection)->mutex);
 			(*connection)->hash_name = hash_name;
 			g_hash_table_insert (sftp_connection_table, hash_name, *connection);
@@ -1229,7 +1435,6 @@ sftp_get_connection (SftpConnection **connection, const GnomeVFSURI *uri)
 		res = GNOME_VFS_OK;
 	}
 
- bail:
 	G_UNLOCK (sftp_connection_table);
 
 	return res;
@@ -1500,7 +1705,7 @@ do_open (GnomeVFSMethod        *method,
 
 	buffer_write_gint32 (&msg, sftp_mode);
 
-	memset (&info, 0, sizeof (GnomeVFSFileInfo));
+	bzero (&info, sizeof (GnomeVFSFileInfo));
 	buffer_write_file_info (&msg, &info, GNOME_VFS_SET_FILE_INFO_NONE);
 
 	buffer_send (&msg, conn->out_fd);
@@ -1583,7 +1788,7 @@ do_create (GnomeVFSMethod        *method,
 
 	buffer_write_gint32 (&msg, ssh_mode);
 
-	memset (&info, 0, sizeof (GnomeVFSFileInfo));
+	bzero (&info, sizeof (GnomeVFSFileInfo));
 	info.permissions = perm;
 	buffer_write_file_info (&msg, &info, GNOME_VFS_SET_FILE_INFO_PERMISSIONS);
 
@@ -2279,8 +2484,7 @@ do_read_directory (GnomeVFSMethod       *method,
 			while (handle->info_write_ptr + count > handle->info_alloc) {
 				handle->info_alloc *= 2;
 				handle->info = g_renew (GnomeVFSFileInfo, handle->info, handle->info_alloc);
-				memset (&(handle->info[handle->info_write_ptr]),
-					0,
+				bzero (&(handle->info[handle->info_write_ptr]),
 				       sizeof (GnomeVFSFileInfo) *
 				       (handle->info_alloc - handle->info_write_ptr));
 			}
@@ -2360,7 +2564,7 @@ do_make_directory (GnomeVFSMethod  *method,
 	id = sftp_connection_get_id (conn);
 
 	path = gnome_vfs_unescape_string (gnome_vfs_uri_get_path (uri), NULL);
-	memset (&info, 0, sizeof (GnomeVFSFileInfo));
+	bzero (&info, sizeof (GnomeVFSFileInfo));
 	iobuf_send_string_request_with_file_info (conn->out_fd, id, SSH2_FXP_MKDIR,
 						  path, strlen (path), &info,
 						  GNOME_VFS_SET_FILE_INFO_NONE);
